@@ -1,5 +1,5 @@
 import dbus
-import six
+import os
 import shutil
 import re
 import tempfile
@@ -8,10 +8,9 @@ import unittest
 from bytesize import bytesize
 from collections import namedtuple
 from contextlib import contextmanager
-from distutils.version import LooseVersion
+from packaging.version import Version
 
 import udiskstestcase
-from udiskstestcase import unstable_test
 
 
 Device = namedtuple('Device', ['obj', 'obj_path', 'path', 'name', 'size'])
@@ -23,7 +22,7 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
     @classmethod
     def setUpClass(cls):
         udiskstestcase.UdisksTestCase.setUpClass()
-        if not cls.check_module_loaded('BTRFS'):
+        if not cls.check_module_loaded('btrfs'):
             udiskstestcase.UdisksTestCase.tearDownClass()
             raise unittest.SkipTest('Udisks module for btrfs tests not loaded, skipping.')
 
@@ -65,7 +64,14 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
         m = re.search(r'[Bb]trfs.* v([\d\.]+)', out)
         if not m or len(m.groups()) != 1:
             raise RuntimeError('Failed to determine btrfs version from: %s' % out)
-        return LooseVersion(m.groups()[0])
+        return Version(m.groups()[0])
+
+    def test__manager_interface(self):
+        '''Test for module D-Bus Manager interface presence'''
+
+        manager = self.get_object('/Manager')
+        intro_data = manager.Introspect(self.no_options, dbus_interface='org.freedesktop.DBus.Introspectable')
+        self.assertIn('interface name="%s.Manager.BTRFS"' % self.iface_prefix, intro_data)
 
     def test_create(self):
         dev = self._get_devices(1)[0]
@@ -77,8 +83,6 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
                              self.no_options,
                              dbus_interface=self.iface_prefix + '.Manager.BTRFS')
 
-        self.write_file("/sys/block/%s/uevent" % dev.name, "change\n")
-
         # check filesystem type
         usage = self.get_property(dev.obj, '.Block', 'IdUsage')
         usage.assertEqual('filesystem')
@@ -106,7 +110,7 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
             sys_size = bytesize.Size(m.group(1))
             self.assertEqual(sys_size.convert_to(bytesize.B), dev.size)
 
-    @unstable_test
+    @udiskstestcase.tag_test(udiskstestcase.TestTags.UNSTABLE)
     def test_create_raid(self):
         devs = self._get_devices(2)
 
@@ -116,8 +120,8 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
         manager = self.get_object('/Manager')
 
         # invalid raid level
-        msg = '[uU]nknown profile raidN'
-        with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+        msg = '[uU]nknown .* profile raidN'
+        with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
             manager.CreateVolume([dev.obj_path for dev in devs],
                                  'test_raidN', 'raidN', 'raidN',
                                  self.no_options,
@@ -159,7 +163,7 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
     def test_subvolumes(self):
 
         btrfs_version = self._get_btrfs_version()
-        if btrfs_version == LooseVersion('4.13.2'):
+        if btrfs_version == Version('4.13.2'):
             self.skipTest('subvolumes list is broken with btrfs-progs v4.13.2')
 
         dev = self._get_devices(1)[0]
@@ -170,14 +174,13 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
                              'test_subvols', 'single', 'single',
                              self.no_options,
                              dbus_interface=self.iface_prefix + '.Manager.BTRFS')
-        self.write_file("/sys/block/%s/uevent" % dev.name, "change\n")
 
         fstype = self.get_property(dev.obj, '.Block', 'IdType')
         fstype.assertEqual('btrfs')
 
         # not mounted, should fail
         msg = 'org.freedesktop.UDisks2.Error.NotMounted: Volume not mounted'
-        with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+        with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
             dev.obj.CreateSubvolume('test_sub1', self.no_options,
                                     dbus_interface=self.iface_prefix + '.Filesystem.BTRFS')
 
@@ -216,8 +219,6 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
                              'test_add_remove', 'single', 'single',
                              self.no_options,
                              dbus_interface=self.iface_prefix + '.Manager.BTRFS')
-        with open("/sys/block/%s/uevent" % dev1.name, "w") as f:
-            f.write("change\n")
 
         fstype = self.get_property(dev1.obj, '.Block', 'IdType')
         fstype.assertEqual('btrfs')
@@ -225,7 +226,7 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
         # shouldn't be possible to remove only device
         with self._temp_mount(dev1.path):
             msg = 'unable to remove the only writeable device'
-            with six.assertRaisesRegex(self, dbus.exceptions.DBusException, msg):
+            with self.assertRaisesRegex(dbus.exceptions.DBusException, msg):
                 dev1.obj.RemoveDevice(dev1.obj_path, self.no_options,
                                       dbus_interface=self.iface_prefix + '.Filesystem.BTRFS')
 
@@ -233,8 +234,6 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
         with self._temp_mount(dev1.path):
             dev1.obj.AddDevice(dev2.obj_path, self.no_options,
                                dbus_interface=self.iface_prefix + '.Filesystem.BTRFS')
-        with open("/sys/block/%s/uevent" % dev2.name, "w") as f:
-            f.write("change\n")
 
         # check filesystem type of the new device
         fstype = self.get_property(dev2.obj, '.Block', 'IdType')
@@ -279,7 +278,7 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
                              'test_snapshot', 'single', 'single',
                              self.no_options,
                              dbus_interface=self.iface_prefix + '.Manager.BTRFS')
-        self.write_file("/sys/block/%s/uevent" % dev.name, "change\n")
+
         fstype = self.get_property(dev.obj, '.Block', 'IdType')
         fstype.assertEqual('btrfs')
 
@@ -311,7 +310,7 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
                              'test_snapshot', 'single', 'single',
                              self.no_options,
                              dbus_interface=self.iface_prefix + '.Manager.BTRFS')
-        self.write_file("/sys/block/%s/uevent" % dev.name, "change\n")
+
         fstype = self.get_property(dev.obj, '.Block', 'IdType')
         fstype.assertEqual('btrfs')
 
@@ -334,17 +333,75 @@ class UdisksBtrfsTest(udiskstestcase.UdisksTestCase):
                              'test_label', 'single', 'single',
                              self.no_options,
                              dbus_interface=self.iface_prefix + '.Manager.BTRFS')
-        self.write_file("/sys/block/%s/uevent" % dev.name, "change\n")
+
         fstype = self.get_property(dev.obj, '.Block', 'IdType')
         fstype.assertEqual('btrfs')
 
         dev.obj.SetLabel('new_label', self.no_options,
                          dbus_interface=self.iface_prefix + '.Filesystem.BTRFS')
 
-        self.write_file("/sys/block/%s/uevent" % dev.name, "change\n")
-        self.udev_settle()
         dbus_label = self.get_property(dev.obj, '.Filesystem.BTRFS', 'label')
         dbus_label.assertEqual('new_label')
 
         _ret, sys_label = self.run_command('lsblk -d -no LABEL %s' % dev.path)
         self.assertEqual(sys_label, 'new_label')
+
+    def test_subvolume_mount(self):
+        dev = self._get_devices(1)[0]
+        self.addCleanup(self._clean_format, dev.obj)
+
+        manager = self.get_object('/Manager')
+        manager.CreateVolume([dev.obj_path],
+                             'test_snapshot', 'single', 'single',
+                             self.no_options,
+                             dbus_interface=self.iface_prefix + '.Manager.BTRFS')
+
+        fstype = self.get_property(dev.obj, '.Block', 'IdType')
+        fstype.assertEqual('btrfs')
+
+        with self._temp_mount(dev.path):
+            # create a subvolume
+            dev.obj.CreateSubvolume('test_sub1', self.no_options,
+                                    dbus_interface=self.iface_prefix + '.Filesystem.BTRFS')
+
+        self.addCleanup(self.try_unmount, dev.path)
+        opts = dbus.Dictionary({'options': 'subvol=/test_sub1'},
+                               signature=dbus.Signature('sv'))
+        mnt_path = dev.obj.Mount(opts, dbus_interface=self.iface_prefix + '.Filesystem')
+
+        # system mountpoint
+        self.assertTrue(os.path.ismount(mnt_path))
+        _ret, out = self.run_command('mount | grep %s' % dev.path)
+        self.assertIn(mnt_path, out)
+        self.assertIn('subvol=/test_sub1', out)
+
+        # dbus mountpoint
+        dbus_mounts = self.get_property(dev.obj, '.Filesystem', 'MountPoints')
+        dbus_mounts.assertLen(1)  # just one mountpoint
+        dbus_mnt = self.ay_to_str(dbus_mounts.value[0])  # mountpoints are arrays of bytes
+        self.assertEqual(dbus_mnt, mnt_path)
+
+    def test_default_subvolume_id(self):
+        dev = self._get_devices(1)[0]
+        self.addCleanup(self._clean_format, dev.obj)
+
+        manager = self.get_object('/Manager')
+        manager.CreateVolume([dev.obj_path],
+                             'test_subvolume_id', 'single', 'single',
+                             self.no_options,
+                             dbus_interface=self.iface_prefix + '.Manager.BTRFS')
+
+        fstype = self.get_property(dev.obj, '.Block', 'IdType')
+        fstype.assertEqual('btrfs')
+
+        with self._temp_mount(dev.path):
+            default_id = dev.obj.GetDefaultSubvolumeID(self.no_options,
+                             dbus_interface=self.iface_prefix + '.Filesystem.BTRFS')
+            self.assertEqual(default_id, 5)
+
+            dev.obj.SetDefaultSubvolumeID(dbus.UInt32(5), self.no_options,
+                                          dbus_interface=self.iface_prefix + '.Filesystem.BTRFS')
+
+            default_id = dev.obj.GetDefaultSubvolumeID(self.no_options,
+                             dbus_interface=self.iface_prefix + '.Filesystem.BTRFS')
+            self.assertEqual(default_id, 5)
