@@ -38,6 +38,7 @@
 #include "udisksmountmonitor.h"
 #include "udiskslinuxdevice.h"
 #include "udisksthreadedjob.h"
+#include "udiskssimplejob.h"
 
 /**
  * SECTION:udiskslinuxswapspace
@@ -126,6 +127,7 @@ udisks_linux_swapspace_update (UDisksLinuxSwapspace   *swapspace,
     active = TRUE;
   udisks_swapspace_set_active (UDISKS_SWAPSPACE (swapspace), active);
 
+  g_dbus_interface_skeleton_flush (G_DBUS_INTERFACE_SKELETON (swapspace));
   g_object_unref (device);
 }
 
@@ -137,13 +139,11 @@ start_job_func (UDisksThreadedJob  *job,
                 gpointer            user_data,
                 GError            **error)
 {
-
-  UDisksObject *object = NULL;
-  UDisksBlock *block = NULL;
-  gchar *device = NULL;
+  UDisksObject *object = UDISKS_OBJECT (user_data);
+  UDisksBlock *block;
+  gchar *device;
   gboolean ret = FALSE;
 
-  object = UDISKS_OBJECT (g_dbus_interface_get_object (G_DBUS_INTERFACE (user_data)));
   block = udisks_object_get_block (object);
   device = udisks_block_dup_device (block);
 
@@ -154,25 +154,6 @@ start_job_func (UDisksThreadedJob  *job,
   return ret;
 }
 
-static void
-swapspace_start_on_job_completed (UDisksJob   *job,
-                                  gboolean     success,
-                                  const gchar *message,
-                                  gpointer     user_data)
-{
-  GDBusMethodInvocation *invocation = G_DBUS_METHOD_INVOCATION (user_data);
-  UDisksSwapspace *swapspace;
-  swapspace = UDISKS_SWAPSPACE (g_dbus_method_invocation_get_user_data (invocation));
-  if (success)
-    udisks_swapspace_complete_start (swapspace, invocation);
-  else
-    g_dbus_method_invocation_return_error (invocation,
-                                           UDISKS_ERROR,
-                                           UDISKS_ERROR_FAILED,
-                                           "Error activating swap: %s",
-                                           message);
-}
-
 static gboolean
 handle_start (UDisksSwapspace        *swapspace,
               GDBusMethodInvocation  *invocation,
@@ -180,10 +161,8 @@ handle_start (UDisksSwapspace        *swapspace,
 {
   UDisksObject *object;
   UDisksDaemon *daemon;
-  UDisksThreadedJob *job;
   GError *error = NULL;
   uid_t caller_uid;
-  gid_t caller_gid;
 
   object = udisks_daemon_util_dup_object (swapspace, &error);
   if (object == NULL)
@@ -194,17 +173,13 @@ handle_start (UDisksSwapspace        *swapspace,
 
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
 
-  error = NULL;
   if (!udisks_daemon_util_get_caller_uid_sync (daemon,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                &error))
     {
-      g_dbus_method_invocation_return_gerror (invocation, error);
-      g_clear_error (&error);
+      g_dbus_method_invocation_take_error (invocation, error);
       goto out;
     }
 
@@ -222,19 +197,29 @@ handle_start (UDisksSwapspace        *swapspace,
                                                     invocation))
     goto out;
 
-  job = UDISKS_THREADED_JOB (udisks_daemon_launch_threaded_job (daemon,
-                                                                object,
-                                                                "swapspace-start",
-                                                                caller_uid,
-                                                                start_job_func,
-                                                                g_object_ref (swapspace),
-                                                                g_object_unref,
-                                                                NULL)); /* cancellable */
-  g_signal_connect (job,
-                    "completed",
-                    G_CALLBACK (swapspace_start_on_job_completed),
-                    invocation);
-  udisks_threaded_job_start (job);
+  if (!udisks_daemon_launch_threaded_job_sync (daemon,
+                                               object,
+                                               "swapspace-start",
+                                               caller_uid,
+                                               FALSE,
+                                               start_job_func,
+                                               object,
+                                               NULL, /* user_data_free_func */
+                                               NULL, /* cancellable */
+                                               &error))
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_FAILED,
+                                             "Error activating swap: %s",
+                                             error->message);
+      g_clear_error (&error);
+      goto out;
+    }
+
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
+  udisks_swapspace_complete_start (swapspace, invocation);
 
  out:
   g_clear_object (&object);
@@ -248,13 +233,11 @@ stop_job_func (UDisksThreadedJob  *job,
                gpointer            user_data,
                GError            **error)
 {
-
-  UDisksObject *object = NULL;
-  UDisksBlock *block = NULL;
-  gchar *device = NULL;
+  UDisksObject *object = UDISKS_OBJECT (user_data);
+  UDisksBlock *block;
+  gchar *device;
   gboolean ret = FALSE;
 
-  object = UDISKS_OBJECT (g_dbus_interface_get_object (G_DBUS_INTERFACE (user_data)));
   block = udisks_object_get_block (object);
   device = udisks_block_dup_device (block);
 
@@ -265,25 +248,6 @@ stop_job_func (UDisksThreadedJob  *job,
   return ret;
 }
 
-static void
-swapspace_stop_on_job_completed (UDisksJob   *job,
-                                 gboolean     success,
-                                 const gchar *message,
-                                 gpointer     user_data)
-{
-  GDBusMethodInvocation *invocation = G_DBUS_METHOD_INVOCATION (user_data);
-  UDisksSwapspace *swapspace;
-  swapspace = UDISKS_SWAPSPACE (g_dbus_method_invocation_get_user_data (invocation));
-  if (success)
-    udisks_swapspace_complete_start (swapspace, invocation);
-  else
-    g_dbus_method_invocation_return_error (invocation,
-                                           UDISKS_ERROR,
-                                           UDISKS_ERROR_FAILED,
-                                           "Error deactivating swap: %s",
-                                           message);
-}
-
 static gboolean
 handle_stop (UDisksSwapspace        *swapspace,
              GDBusMethodInvocation  *invocation,
@@ -291,25 +255,25 @@ handle_stop (UDisksSwapspace        *swapspace,
 {
   UDisksObject *object;
   UDisksDaemon *daemon;
-  UDisksThreadedJob *job;
   uid_t caller_uid;
-  gid_t caller_gid;
   GError *error = NULL;
 
-  object = UDISKS_OBJECT (g_dbus_interface_get_object (G_DBUS_INTERFACE (swapspace)));
+  object = udisks_daemon_util_dup_object (swapspace, &error);
+  if (object == NULL)
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      goto out;
+    }
+
   daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
 
-  error = NULL;
   if (!udisks_daemon_util_get_caller_uid_sync (daemon,
                                                invocation,
                                                NULL /* GCancellable */,
                                                &caller_uid,
-                                               &caller_gid,
-                                               NULL,
                                                &error))
     {
-      g_dbus_method_invocation_return_gerror (invocation, error);
-      g_clear_error (&error);
+      g_dbus_method_invocation_take_error (invocation, error);
       goto out;
     }
 
@@ -332,22 +296,206 @@ handle_stop (UDisksSwapspace        *swapspace,
                                                     invocation))
     goto out;
 
-  job = UDISKS_THREADED_JOB (udisks_daemon_launch_threaded_job (daemon,
-                                                                object,
-                                                                "swapspace-stop",
-                                                                caller_uid,
-                                                                stop_job_func,
-                                                                g_object_ref (swapspace),
-                                                                g_object_unref,
-                                                                NULL)); /* cancellable */
+  if (!udisks_daemon_launch_threaded_job_sync (daemon,
+                                               object,
+                                               "swapspace-stop",
+                                               caller_uid,
+                                               FALSE,
+                                               stop_job_func,
+                                               object,
+                                               NULL, /* user_data_free_func */
+                                               NULL, /* cancellable */
+                                               &error))
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_FAILED,
+                                             "Error deactivating swap: %s",
+                                             error->message);
+      g_clear_error (&error);
+      goto out;
+    }
 
-  g_signal_connect (job,
-                    "completed",
-                    G_CALLBACK (swapspace_stop_on_job_completed),
-                    invocation);
-  udisks_threaded_job_start (job);
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
+  udisks_swapspace_complete_stop (swapspace, invocation);
 
  out:
+  g_clear_object (&object);
+  return TRUE;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+static gboolean
+handle_set_label (UDisksSwapspace        *swapspace,
+                  GDBusMethodInvocation  *invocation,
+                  const gchar            *label,
+                  GVariant               *options)
+{
+  UDisksObject *object;
+  UDisksDaemon *daemon;
+  UDisksBaseJob *job = NULL;
+  GError *error = NULL;
+  UDisksBlock *block = NULL;
+  uid_t caller_uid;
+
+  object = udisks_daemon_util_dup_object (swapspace, &error);
+  if (object == NULL)
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      goto out;
+    }
+
+  daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  block = udisks_object_peek_block (object);
+
+  if (!udisks_daemon_util_get_caller_uid_sync (daemon,
+                                               invocation,
+                                               NULL /* GCancellable */,
+                                               &caller_uid,
+                                               &error))
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      goto out;
+    }
+
+  if (!udisks_daemon_util_check_authorization_sync (daemon,
+                                                    object,
+                                                    "org.freedesktop.udisks2.manage-swapspace",
+                                                    options,
+                                                    /* Translators: Shown in authentication dialog when the user
+                                                     * requests setting label of a swap device.
+                                                     *
+                                                     * Do not translate $(drive), it's a placeholder and
+                                                     * will be replaced by the name of the drive/device in question
+                                                     */
+                                                    N_("Authentication is required to set swapspace label on $(drive)"),
+                                                    invocation))
+    goto out;
+
+  job = udisks_daemon_launch_simple_job (daemon,
+                                         UDISKS_OBJECT (object),
+                                         "swapspace-modify",
+                                         caller_uid,
+                                         FALSE,
+                                         NULL);
+  if (job == NULL)
+    {
+      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
+                                             "Failed to create a job object");
+      goto out;
+    }
+
+  if (!bd_swap_set_label (udisks_block_get_device (block),
+                          label,
+                          &error))
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_FAILED,
+                                             "Error taking setting label on %s: %s",
+                                             udisks_block_get_device (block),
+                                             error->message);
+      udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
+      g_clear_error (&error);
+      goto out;
+    }
+
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
+  udisks_swapspace_complete_set_label (swapspace, invocation);
+  udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
+
+ out:
+  g_clear_object (&object);
+  return TRUE;
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
+static gboolean
+handle_set_uuid (UDisksSwapspace       *swapspace,
+                 GDBusMethodInvocation *invocation,
+                 const gchar           *arg_uuid,
+                 GVariant              *arg_options)
+{
+  UDisksObject *object;
+  UDisksDaemon *daemon;
+  UDisksBaseJob *job = NULL;
+  GError *error = NULL;
+  UDisksBlock *block = NULL;
+  uid_t caller_uid;
+
+  object = udisks_daemon_util_dup_object (swapspace, &error);
+  if (object == NULL)
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      goto out;
+    }
+
+  daemon = udisks_linux_block_object_get_daemon (UDISKS_LINUX_BLOCK_OBJECT (object));
+  block = udisks_object_peek_block (object);
+
+  if (!udisks_daemon_util_get_caller_uid_sync (daemon,
+                                               invocation,
+                                               NULL /* GCancellable */,
+                                               &caller_uid,
+                                               &error))
+    {
+      g_dbus_method_invocation_take_error (invocation, error);
+      goto out;
+    }
+
+  if (!udisks_daemon_util_check_authorization_sync (daemon,
+                                                    object,
+                                                    "org.freedesktop.udisks2.manage-swapspace",
+                                                    arg_options,
+                                                    /* Translators: Shown in authentication dialog when the user
+                                                     * requests setting UUID of a swap device.
+                                                     *
+                                                     * Do not translate $(drive), it's a placeholder and
+                                                     * will be replaced by the name of the drive/device in question
+                                                     */
+                                                    N_("Authentication is required to set swapspace UUID on $(drive)"),
+                                                    invocation))
+    goto out;
+
+  job = udisks_daemon_launch_simple_job (daemon,
+                                         UDISKS_OBJECT (object),
+                                         "swapspace-modify",
+                                         caller_uid,
+                                         FALSE,
+                                         NULL);
+  if (job == NULL)
+    {
+      g_dbus_method_invocation_return_error (invocation, UDISKS_ERROR, UDISKS_ERROR_FAILED,
+                                             "Failed to create a job object");
+      goto out;
+    }
+
+  if (!bd_swap_set_uuid (udisks_block_get_device (block),
+                         arg_uuid,
+                         &error))
+    {
+      g_dbus_method_invocation_return_error (invocation,
+                                             UDISKS_ERROR,
+                                             UDISKS_ERROR_FAILED,
+                                             "Error taking setting UUID on %s: %s",
+                                             udisks_block_get_device (block),
+                                             error->message);
+      udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), FALSE, error->message);
+      g_clear_error (&error);
+      goto out;
+    }
+
+  udisks_linux_block_object_trigger_uevent_sync (UDISKS_LINUX_BLOCK_OBJECT (object),
+                                                 UDISKS_DEFAULT_WAIT_TIMEOUT);
+  udisks_swapspace_complete_set_uuid (swapspace, invocation);
+  udisks_simple_job_complete (UDISKS_SIMPLE_JOB (job), TRUE, NULL);
+
+ out:
+  g_clear_object (&object);
   return TRUE;
 }
 
@@ -356,6 +504,8 @@ handle_stop (UDisksSwapspace        *swapspace,
 static void
 swapspace_iface_init (UDisksSwapspaceIface *iface)
 {
-  iface->handle_start  = handle_start;
-  iface->handle_stop   = handle_stop;
+  iface->handle_start     = handle_start;
+  iface->handle_stop      = handle_stop;
+  iface->handle_set_label = handle_set_label;
+  iface->handle_set_uuid  = handle_set_uuid;
 }
